package de.moritzruth.theaterdsl.device

import de.moritzruth.theaterdsl.value.*
import kotlinx.collections.immutable.ImmutableList
import kotlinx.collections.immutable.persistentListOf
import kotlinx.collections.immutable.toImmutableList
import kotlin.math.PI
import kotlin.math.asin
import kotlin.math.min
import kotlin.math.sin
import kotlin.random.Random
import kotlin.time.Duration
import kotlin.time.ExperimentalTime
import kotlin.time.TimeSource

interface DynamicValue<T> {
    fun reset()
    fun getCurrentValue(): T
}

@OptIn(ExperimentalTime::class)
abstract class DoubleDV<T>(private val initialStaticValue: Double) : DynamicValue<T> {
    private sealed interface State {
        data class Static(val value: Double) : State

        data class Fade(val start: Double, val end: Double, val duration: Duration) : State {
            val delta = end - start
        }

        data class Sine(
            val minimum: Double,
            val maximum: Double,
            val b: Double,
            val c: Double
        ) : State {
            companion object {
                fun calculateB(period: Duration) = (2 * PI) / period.inWholeMilliseconds
                fun calculateX(elapsedTime: Duration, b: Double, c: Double) = b * elapsedTime.inWholeMilliseconds + c
            }
        }

        data class Ramp(val period: Duration, val start: Double, val end: Double) : State
        data class TriangleWave(val period: Duration, val start: Double, val end: Double) : State
        data class Step(val steps: ImmutableList<Double>, val interval: Duration, val startIndex: Int) : State
        data class PulseOnce(val rampUpDuration: Duration, val rampDownDuration: Duration, val peakValue: Double, val start: Double, val end: Double) : State
    }

    protected abstract fun toDomain(value: Double): T
    protected abstract fun fromDomain(value: T): Double
    protected abstract val minimumValue: T
    protected abstract val maximumValue: T

    private var stateChangeMark = TimeSource.Monotonic.markNow()
    private var state: State = State.Static(initialStaticValue)
        set(value) {
            field = value
            stateChangeMark = TimeSource.Monotonic.markNow()
        }

    override fun reset() {
        state = State.Static(initialStaticValue)
    }

    override fun getCurrentValue(): T {
        val elapsedTime = stateChangeMark.elapsedNow()

        val double = when (val s = state) {
            is State.Static -> s.value
            is State.Fade -> (min(elapsedTime / s.duration, 1.0) * s.delta + s.start)
            is State.Sine -> {
                val fromZeroToOne = (sin(State.Sine.calculateX(elapsedTime, s.b, s.c)) + 1) / 2
                (fromZeroToOne * (s.maximum - s.minimum) + s.minimum)
            }

            is State.Ramp -> {
                val progress = elapsedTime.inWholeMilliseconds.toDouble() / s.period.inWholeMilliseconds
                val clampedProcess = progress.mod(1.0)

                clampedProcess * (s.end - s.start) + s.start
            }

            is State.TriangleWave -> {
                val progress = elapsedTime.inWholeMilliseconds.toDouble() / s.period.inWholeMilliseconds
                val clampedProcess = progress.mod(1.0)
                1.0
            }

            is State.Step -> {
                val index = (elapsedTime / s.interval).toInt() + s.startIndex
                s.steps[index.mod(s.steps.size)]
            }

            is State.PulseOnce -> {
                if (elapsedTime <= s.rampUpDuration) {
                    val progress = elapsedTime / s.rampUpDuration
                    val delta = s.peakValue - s.start
                    (progress * delta + s.start)
                } else {
                    val progress = min(elapsedTime / s.rampDownDuration, 1.0)
                    val delta = s.peakValue - s.end
                    (s.peakValue - progress * delta)
                }
            }
        }

        return toDomain(double)
    }

    fun static(value: T) {
        state = State.Static(fromDomain(value))
    }

    fun off(duration: Duration = Duration.ZERO) = fade(minimumValue, duration)

    fun fade(end: T, duration: Duration, start: T = getCurrentValue()) {
        state =
            if (duration == Duration.ZERO || duration.isNegative()) State.Static(fromDomain(end))
            else State.Fade(fromDomain(start), fromDomain(end), duration)
    }

    fun sine(period: Duration, minimum: T = minimumValue, maximum: T = maximumValue, start: T = getCurrentValue()) {
        val b = State.Sine.calculateB(period)
        val doubleMinimum = fromDomain(minimum)
        val doubleMaximum = fromDomain(maximum)
        val delta = doubleMaximum - doubleMinimum

        val coercedStart = fromDomain(start).coerceIn(doubleMinimum, fromDomain(maximum))
        val c = asin(2 * ((coercedStart - doubleMinimum) / delta) - 1)
        state = State.Sine(fromDomain(minimum), fromDomain(maximum), b, c)
    }

    fun ramp(period: Duration, start: T, end: T) {
        state = State.Ramp(period, fromDomain(start), fromDomain(end))
    }

    fun rampUp(period: Duration) = ramp(period, minimumValue, maximumValue)
    fun rampDown(period: Duration) = ramp(period, maximumValue, minimumValue)

    fun steps(steps: List<T>, interval: Duration, startIndex: Int = 0) {
        state = State.Step(steps.map(this::fromDomain).toImmutableList(), interval, startIndex)
    }

    fun switch(a: T, b: T, interval: Duration, startAtCurrent: Boolean = true) {
        val startIndex = startAtCurrent
            .takeIf { it }
            ?.let { getCurrentValue() }
            ?.takeIf { it == b }
            ?.let { 1 } ?: 0

        steps(persistentListOf(a, b), interval, startIndex)
    }

    fun pulseOnce(rampUpDuration: Duration, rampDownDuration: Duration, peak: T = maximumValue, end: T = minimumValue, start: T = getCurrentValue()) {
        state = State.PulseOnce(rampUpDuration, rampDownDuration, fromDomain(peak), fromDomain(start), fromDomain(end))
    }
}

class PercentageDV(initialStaticValue: Percentage = 0.percent) : DoubleDV<Percentage>(initialStaticValue.value) {
    override fun fromDomain(value: Percentage): Double = value.value
    override fun toDomain(value: Double): Percentage = Percentage(value)
    override val minimumValue: Percentage = 0.percent
    override val maximumValue: Percentage = 100.percent
}

class AngleDV(initialStaticValue: Angle = 0.degrees, override val maximumValue: Angle = 360.degrees) : DoubleDV<Angle>(initialStaticValue.degrees) {
    override fun fromDomain(value: Angle): Double = value.degrees
    override fun toDomain(value: Double): Angle = Angle(value)
    override val minimumValue: Angle = 0.degrees
}

class ConcreteDV<T>(private val initialStaticValue: T) : DynamicValue<T> {
    sealed interface State<T> {
        class Static<T>(val value: T) : State<T>
    }

    var state: State<T> = State.Static(initialStaticValue)

    override fun reset() {
        state = State.Static(initialStaticValue)
    }

    override fun getCurrentValue(): T = when (val s = state) {
        is State.Static -> s.value
    }

    fun static(value: T) {
        state = State.Static(value)
    }
}

@OptIn(ExperimentalTime::class)
class ColorDV(private val initialStaticValue: Color = Color.WHITE) : DynamicValue<Color> {
    private sealed interface State {
        data class Static(val value: Color) : State
        data class Fade(val start: Color, val end: Color, val duration: Duration) : State {
            val deltaHue = ((end.hue.degrees - start.hue.degrees).mod(360f) + 540f).mod(360f) - 180f
            val deltaSaturation = end.saturation.value - start.saturation.value
            val deltaBrightness = end.brightness.value - start.brightness.value
        }

        data class Cycle(val start: Color, val interval: Duration) : State

        data class FadeRandomAround(val hue: Angle, val deviation: Angle, val interval: Duration) : State
    }

    private var stateChangeMark = TimeSource.Monotonic.markNow()
    private var state: State = State.Static(initialStaticValue)
        set(value) {
            field = value
            stateChangeMark = TimeSource.Monotonic.markNow()
        }

    override fun reset() {
        state = State.Static(initialStaticValue)
    }

    override fun getCurrentValue(): Color {
        val elapsedTime = stateChangeMark.elapsedNow()

        return when (val s = state) {
            is State.Static -> s.value
            is State.Fade -> {
                val progress = min(elapsedTime / s.duration, 1.0)

                Color(
                    hue = Angle((s.start.hue.degrees + s.deltaHue * progress)),
                    saturation = Percentage((s.start.saturation.value + s.deltaSaturation * progress)),
                    brightness = Percentage((s.start.brightness.value + s.deltaBrightness * progress))
                )
            }

            is State.Cycle -> {
                val progress = (elapsedTime / s.interval).mod(1.0)
                s.start.copy(hue = (s.start.hue.degrees + progress * 360.0).degrees.inSingleRotation())
            }

            is State.FadeRandomAround -> {
                val startRandom = Random((elapsedTime / s.interval).toInt() - 1)
                val endRandom = Random((elapsedTime / s.interval).toInt())
                val progress = (elapsedTime / s.interval).mod(1.0)
                val startHue = s.hue.degrees - s.deviation.degrees + startRandom.nextDouble() * s.deviation.degrees * 2
                val endHue = s.hue.degrees - s.deviation.degrees + endRandom.nextDouble() * s.deviation.degrees * 2
                val delta = endHue - startHue

                Color(hue = Angle((startHue + progress * delta)))
            }
        }
    }

    fun static(value: Color) {
        state = State.Static(value)
    }

    fun fade(end: Color, duration: Duration, start: Color = getCurrentValue()) {
        state = State.Fade(start, end, duration)
    }

    fun cycle(interval: Duration, start: Color = getCurrentValue().copy(saturation = 100.percent, brightness = 100.percent)) {
        state = State.Cycle(start, interval)
    }

    fun fadeRandomAround(hue: Angle, deviation: Angle, interval: Duration) {
        state = State.FadeRandomAround(hue, deviation, interval)
    }
}